import subprocess
import os

env = os.environ.copy()
env["KGSTEWARD_ROOT_DIR"] = os.getcwd()

def run_cmd( cmd: list[str], env: dict[ str, str ] = env ):
    return subprocess.run(
        cmd,
        capture_output = True,
        text = True,
        cwd = env["KGSTEWARD_ROOT_DIR"],
        env = env
    )



from dumper import dump
import re
from .common import *
from .generic import GenericClient

class OxigraphClient( GenericClient ):

    def __init__( self, oxigraph_url ):
        self.oxigraph_url = oxigraph_url
        super().__init__(  
            oxigraph_url + "/query", 
            oxigraph_url + "/update",
            oxigraph_url + "/data"
        )

    def ping( self ):
        super().ping()
        #r = http_call({
        #    'method'  : 'GET',
        #           'url'     : self.oxigraph_url + "/#/"
        #       },[ 200 ] )
        #ignore = super().ping()
        #print( "OK" )

    def rewrite_repository( self, server_config_filename ) :
        self.sparql_update( "DROP ALL")

    def free_access( self ) :
        print_warn( "Not yet implemented: FusekiClient.free_access()" );

    def sparql_query( 
        self, sparql, 
        headers = { 
            'Content-Type' : 'application/x-www-form-urlencoded', 
            'Accept'       : 'application/json'
        }, 
        status_code_ok = [ 200 ], 
        echo = True 
    ):
        return super().sparql_query( 
            sparql, 
            headers = headers, 
            status_code_ok = status_code_ok, 
            echo = echo
        )
    
    # on success fuseki return 200, not 204 !!!
    def sparql_update( 
        self, 
        sparql,
        headers = { 'Content-Type' : 'application/x-www-form-urlencoded' },
        status_code_ok = [ 204 ], 
        echo = True
    ):
        super().sparql_update( 
            sparql, 
            headers = headers,
            status_code_ok = status_code_ok, 
            echo = echo 
        )

    def sparql_query_to_tsv( self, sparql, status_code_ok = [200], echo=True) :
        return self.sparql_query(
            sparql,
            accept='text/tab-separated-values',
            status_code_ok=status_code_ok,
            echo=echo,
        )
        
    def graphdb_call( self, request_args, status_code_ok = [ 200 ], echo = True ) :
        print_warn( "Not yet implemented: FusekiClient.graphdb_call()" )

    def validate_sparql_query( self, sparql, echo = False ):
        print_warn( "Not yet implemented: FusekiClient.validate_sparql_query ()" )
        r = http_call({
            'method'  : 'GET',
            'url'     : self.graphdb_url + "/repositories/" + self.repository_id,
            'headers' : {
                'Accept'        : 'text/tab-separated-values',
                'Authorization' : self.authorization
            },
            'params'  : {
                'query'   : sparql,
                "infer"   : True,
                "timeout" : 5       # 5 seems to solve a HTTP "problem" observed with a timout of 1 s ?!?
            }
        }, [ 503, 500, 400, 200 ], echo )
        if r.status_code == 503 :
            time.sleep( 1 )
            print( "\t" + "query timed out" )
        elif r.status_code == 500 : # is returned by GraphDB on timeout of SPARQL queries with a SERVICE clause ?!?
            time.sleep( 1 )
            print( "\t" + "unknown error, maybe timeout" )
        elif r.status_code == 400 :
            raise RuntimeError( "Suspected SPARQL syntax error:\n" + sparql )
        else : #  r.status_code == 200 :
            n = r.text.count( "\n" )
            if n == 0 :
                print( "\t" + "!!! empty results !!!" )
            else :
                print( "\t" + str( n ) + " lines returned" )

    def load_from_url( self, url, context, tmpdir = "/tmp", echo = True ):
        if re.search( r"\.ttl$", url ):
            self.sparql_update( f"LOAD SILENT <{url}> INTO GRAPH <{context}>", echo = echo )
        elif re.search( r"\.(gz|bz2|xz)$", url  ):
            filename = tmpdir + "/" + url.split('/')[-1]
            if echo:
                report( 'write file', filename )
            download_file( url, filename )
            self.load_from_file( filename, context, echo = echo )
        else:
            stop_error( "Cannot handle this type of link with Fuseki: " + url )



# This is a very minimalistic/generic SPARQL client
# that might work on a real triplestore (maybe?)

import subprocess
import urllib

# How to dump a context from the command line:
# curl -G --data-urlencode "graph=http://example.org/context/ReconX_schema" http://localhost:7200/repositories/ReconXKG2/rdf-graphs/service

from dumper  import dump
from .common import *

class GenericClient():

    def __init__( self, endpoint_query, endpoint_update, endpoint_store ):
        self.endpoint_query  = endpoint_query
        self.endpoint_update = endpoint_update
        self.endpoint_store  = endpoint_store
        self.cookies         = None
        self.headers         = None

    def rewrite_repository( self, arg ):
        """ Create an empty repository or empty an existing one """
        raise Exception( "Abstract method called: rewrite_repository()" )
    
    def get_endpoint_query( self ):
        return self.endpoint_query

    def get_endpoint_update( self ):
        return self.endpoint_update
    
    def sparql_query( self, sparql, status_code_ok = [ 200 ], echo = True, timeout = None ): 
        """ Run a sparql query and return the response with the data in JSON format. """
        raise Exception( "Abstract method called: sparql_query()" )
     
    def sparql_update( self, sparql, status_code_ok = [ 204 ], echo = True ):
        """ Run a sparql update, returns nothing """
        raise Exception( "Abstract method called: sparql_update()" )
    
    def list_context( self, echo = True ):
        """ Run the actual list of contexts """
        raise Exception( "Abstract method called: list_context()" )

    def drop_context( self, context ):
        """ Drop a context """
        raise Exception( "Abstract method called: drop_context()" )
    
    def dump_context( self, context, echo = True ):
        r = self.sparql_query( f"""
SELECT ?s ?p ?o 
WHERE{{ 
    GRAPH <{context}> {{ 
        ?s ?p ?o 
    }}
}}"""
)
        return sparql_result_to_table( r )

    def validate_sparql_query( self, sparql, echo = False, timeout = None ):
        """ verify that at least the query returns at least one row of data, or timeout """
        r = self.sparql_query( sparql, echo = False, timeout = timeout )
        if r is None:
            if timeout is not None:
                time.sleep( 1 ) # print_warn( "Query timed out" ) already printed
            else:
                 print( colored( sparql, "green" ))
                 stop_error( "Unknown error!" ) # 
        elif r.status_code == 200 :
            h, v = sparql_result_to_table( r )
            if len( v ) == 0 :
                print_warn( "empty result" )
            else :
                report( "#row", str( len( v )))
        else :
            print( colored( sparql, "green" ))
            stop_error( "Unexpected status code: " + str( r.status_code ))

    def load_from_file( self, file, context, headers = {}, echo = True ):
        """ use graph store protocol """
        if echo:
            report( 'load file', file )
        with any_open( file, 'rb' ) as f: # NB any_open takes care of decompression
            http_call(
                {
                    'method'  : 'POST',
                    'url'     : self.endpoint_store + "?graph=" + urllib.parse.quote_plus( context ),
                    'headers' : {
                        **headers,
                        'Content-Type' : guess_mime_type( file )
                    },
                    'cookies' : self.cookies,
                    'data'    : f
                },
                [ 200, 201, 204 ], # fuseki 200, GraphDB 204
                echo
            )
 
    def _flush_buf( self, context, data, headers = {}, echo = True ):
        args = {
            'method'  : 'POST',
            'url'     : self.endpoint_store + "?graph=" + urllib.parse.quote_plus( context ),
            'headers' : {
                **headers,
                'Content-Type' : 'text/plain', # FIXME: verify encoding as UTF-8
            },
            'data'    : data
        }
        if hasattr( self, "cookies" ):
            args['cookies'] = self.cookies
        http_call( args, [ 200, 201, 204 ], # fuseki 200, GraphDB 204
            echo
    )
        
    def load_from_file_using_riot( self, file, context, headers = {}, echo = True ):
        """ use graph store protocol and riot """
        if echo:
            report( 'load file', file )
        cmd = [ 'riot', file ]
        print( colored( " ".join( cmd ), "cyan" ))
        p = subprocess.Popen( cmd, stdout = subprocess.PIPE, text=True ) # riot returns nt format by default
        buf = []
        size = 0
        count = 0
        for line in p.stdout:
            count += 1
            l = len( line )
            if size + l > 1e8 : # aka 100 mega
                report( "triples so far", str( count ))
                self._flush_buf( context, "".join( buf ), headers )
                buf  = [ line ] # rewrite
                size = l        # rewrite
            else:
                buf.append( line )
                size += l
        if size > 0 :
            report( "triples so far", str( count ))
            self._flush_buf( context, "".join( buf ), headers )



from dumper import dump
import re
import rdflib
import urllib

from .common import *
from .generic import GenericClient

# FIXME: implement curl -XPOST http://localhost:3030/$/compact/TEST?deleteOld=true
# FIXME: test env variable JVM_ARGS

class FusekiClient( GenericClient ):

    def __init__( self, location, repository, config_file, username = None, password = None  ):
        g = rdflib.Graph()
        try:
            g.parse( config_file )
        except:
            stop_error( "Cannot parse file: " + config_file )
        query = """PREFIX rdf: <http://www.w3.org/1999/02/22-rdf-syntax-ns#>
PREFIX fuseki: <http://jena.apache.org/fuseki#>
SELECT ?query ?update ?store
WHERE{
    [] rdf:type fuseki:Service ;
    fuseki:name '""" + repository + """' ;
    fuseki:endpoint [ 
        fuseki:operation fuseki:query ;
        fuseki:name ?query
    ] ;
    fuseki:endpoint [
        fuseki:operation fuseki:update ;
        fuseki:name ?update
    ] ;
    fuseki:endpoint [ 
        fuseki:operation fuseki:gsp-rw ; 
        fuseki:name ?store
    ]
}"""
        qres = g.query( query )
        for row in qres:
            dump( row )
            super().__init__(  
                location + "/" + repository + "/" + row.query, 
                location + "/" + repository + "/" + row.update,
                location + "/" + repository + "/" + row.store
            )
        self.repository = repository 
        self.location   = location
        print_task( "contacting server" )
        try:
            r = http_call({
                'method'  : 'GET',
                'url'     : location + "/$/ping"
            },[ 200 ] )
        except:
            stop_error( "Cannot contact fuseki at location: " + location ) 
        if username is not None :
            try:
                r = http_call({
                    'method'  : 'GET',
                    'url'     : self.fuseki_url + "/$/server",
                    'auth'    : ( username, password )
                },[ 200 ] ) # returns 401 if authentification fail
                self.cookies = requests.utils.dict_from_cookiejar( r.cookies )
            except:
                stop_error( "Authentication to fuseki server failed: + self.graphdb_url" )
        print( self.endpoint_query )
        print( self.endpoint_update )
        print( self.endpoint_store )

    def rewrite_repository( self, server_config_filename ) :
        self.sparql_update( "DROP ALL")
# FIXME: this is n'importe quoi!
#        http_call({
#            'method'  : 'DELETE',
#            'url'     : self.endpoint_store + "?graph=" + urllib.parse.quote_plus( context ),
#            # 'headers' : { **self.headers, **headers }
#        }, [ 204, 404 ] ) # 204: deletion succeful, 404: graph did not exist

    def free_access( self ) :
        print_warn( "Not yet implemented: FusekiClient.free_access()" );

    def sparql_query( self, sparql, status_code_ok = [ 200 ], echo = True, timeout = None ):
        if echo :
            print_strip( sparql.replace( "\t", "    " ), color = "green" )
        headers = {
            'Accept' : 'application/json', 
            'Content-Type': 'application/x-www-form-urlencoded' 
        }
        params = { 'query' : sparql }
        if timeout is not None:
            if not 503 in status_code_ok: # returned by "normal" GraphDB timeout
                status_code_ok.append( 503 ) 
            if not 500 in status_code_ok: # returned by "service" GraphDB timeout
                status_code_ok.append( 500 )
            params["timeout"] = timeout
        r = http_call(
            {
                'method'  : 'POST',  # allows for big query string
                'url'     : self.endpoint_query,
                'headers' : headers,
                'params'  : params,
                'cookies' : self.cookies
            },
            status_code_ok,
        )
        if timeout is not None:
            if r.status_code == 503 :
                time.sleep( 1 )
                print_warn( "query timed out" )
                return None
            elif r.status_code == 500 : # is returned by GraphDB on timeout of SPARQL queries with a SERVICE clause ?!?
                time.sleep( 1 )
                print_warn( "status", "unknown error, possibly timeout" )
                return None

        return r
    
    def sparql_update( self, sparql, status_code_ok = [ 200 ], echo = True ):
        if echo:
            print_strip( sparql.replace( "\t", "    " ), color = "green" )
        http_call(
            {
                'method'  : 'POST',
                'url'     : self.endpoint_update,
                'headers' : { 'Content-Type': 'application/x-www-form-urlencoded' }, 
                'cookies' : self.cookies,
                'params'  : { 'update': sparql },
            }, 
            status_code_ok,
            echo
        )

    def list_context( self, echo = True ):
        r = self.sparql_query( "SELECT DISTINCT ?g WHERE{ GRAPH ?g {}}", echo = echo )
        contexts = set()
        for rec in r.json()["results"]["bindings"]:
            if "g" in rec:
                contexts.add( rec["g"]["value"] )
        return contexts

    def drop_context( self, context, echo = True ): # FIXME: might be faster than DROP GRAPH <context>, echo is ignored
        http_call({
            'method'  : 'DELETE',
            'url'     : self.endpoint_store + "?graph=" + urllib.parse.quote_plus( context ),
            # 'headers' : { **self.headers, **headers }
            'cookies' : self.cookies,
        }, [ 204, 404 ] ) # 204: deletion succeful, 404: graph did not exist

#    def dump_context( self, context, status_code_ok = [ 200 ], echo = True ):
#        return http_call({
#            'method'  : 'GET',
#            'url'     : self.endpoint_store + "?graph=" + urllib.parse.quote_plus( context ),
#            'cookies' : self.cookies,
##            'headers' : { 'Accept': 'text/plain' },
#        }, status_code_ok, echo )

    def fuseki_compress_tdb2( self ):
        r = http_call({
            'method'  : 'POST',
            'url'     : self.location + "/$/compact/" + self.repository + "?deleteOld=true",
            'cookies' : self.cookies,
        }, [ 200 ] )

    def load_from_url( self, url, context, tmpdir = "/tmp", echo = True ): # FIXME: this code is never called
        if re.search( r"\.ttl$", url ):
            self.sparql_update( f"LOAD SILENT <{url}> INTO GRAPH <{context}>", echo = echo )
        elif re.search( r"\.(gz|bz2|xz)$", url  ):
            filename = tmpdir + "/" + url.split('/')[-1]
            if echo:
                report( 'write file', filename )
            download_file( url, filename )
            self.load_from_file( filename, context, echo = echo )
        else:
            stop_error( "Cannot handle this type of link with Fuseki: " + url )



# The code below was strongly inspired by
#    https://github.com/python/cpython/blob/3.12/Lib/http/server.py
# eventhough aI don't understand all the technical details

import http.server
import time
import multiprocessing

class MyServer( http.server.ThreadingHTTPServer ):

    def __init__( self, addr, HandlerClass, directory ): # an ad hoc constructor with the directory parameter
        super().__init__( addr, HandlerClass )
        self.directory = directory

    def finish_request(self, request, client_address):
        self.RequestHandlerClass( request, client_address, self, directory=self.directory )

def _expose_directory( directory, port = 8000 ):
    HandlerClass = http.server.SimpleHTTPRequestHandler
    httpd        = MyServer(( "", port ), HandlerClass, directory )
    httpd.serve_forever()

class LocalFileServer() :
    port      = None
    directory = None
    thread    = None

    def __init__( self, port = 8000 ) :
        self.port = port

    def expose( self, directory ) :
        if( self.thread and self.thread.is_alive() ):
            if( self.directory == directory ):
                return # directory is already exposed
            else :
                self.terminate()
        self.directory = directory
        mp_context = multiprocessing.get_context( 'fork' )
        self.thread = mp_context.Process(
            target=_expose_directory,
            args=( self.directory, self.port ),
            daemon=True
        )
        self.thread.start()
        print( f"# Directory {self.directory} is exposed on http://localhost:{self.port}" )
        time.sleep( 0.2 ) # leaves some time for the seerver to start

    def terminate( self ) :
        if( self.thread and self.thread.is_alive() ):
            self.thread.terminate();
        self.directory = None
        self.thread    = None

def test():
    lfs = LocalFileServer( 8000 )
    lfs.expose( "/Users/mpagni/github.com" )
    time.sleep( 2 )
    lfs.expose( "/Users/mpagni/gitlab.com" )
    time.sleep( 2 )
    lfs.terminate()

# test()





import sys
import json
import jsonschema_markdown
from kgsteward.yamlconfig import KGStewardConf

main_model_schema = KGStewardConf.model_json_schema()
main_model_schema["description"] = "Top level YAML keys"

with open( "kgsteward.schema.json", "w" ) as f:
	f.write( json.dumps( main_model_schema, indent=2 ))

with open( "kgsteward.schema.md", "w" ) as f:
    f.write( """<sup>back to [TOC](../README.md)</sup>

# YAML syntax of kgsteward config file (version 2)

## Preambule
            
* YAML 1.1 syntax is supported. 

* A single YAML extension is supported: `!include <filename>`. 
This directive will insert in place the content of `filename`.
The path of `<filename>` is interpreted with the directory of the parent YAML file as default directory. 
This inclusion mechanism is executed early, before the YAML configuration is validated.  

* Within the YAM config file(s), UNIX environment variables can by referred to using `${...}` syntax. 
Evaluation of these is performed late, i.e. at the time of command execution. 
Hence `${...}` syntax cannot be used in `!include` directive.
The use of UNIX environment variables is recommended to ensure portability of the YAML config file(s).
These variable are usually encoded with uppercase strings.

* In addition to UNIX environment variables, `kgsteward` creates temporary variables reflecting the content of the YAML config file.
For example `${kgsteward_server_brand}` contains the ... server brand, e.g. `graphdb`.
The most useful of these variables is certainly `${kgsteward_dataset_context"}` that contains the IRI of the current target context.
These variable are encoded with lowercase strings.

* The terminology adopted here is a compromise. Different server brands utilise different namings for the same concept. 
For example, 'context' in RDF4J/GraphDB terminology is the same as 'named graph' in RDF/SPARQL terminology.
In this respect, `kgsteward` utilises 'context', because of the too many usages of '[graph](https://en.wikipedia.org/wiki/Graph)'.  

## YAML syntax

The entry point (top level keys) is [KGStewardConf](#kgstewardconf).

""" )
    txt = jsonschema_markdown.generate( main_model_schema, hide_empty_columns = True )
    big_space = "&nbsp;" * 40
    txt = txt.replace( "| Description |",  "|" + big_space + "Description" + big_space + "|" ) # dirty patch to improve table layout
    
    f.write( txt + "\n\n<sup>back to [TOC](../README.md)</sup>" )



# https://stackoverflow.com/questions/58703926/how-do-i-generate-yaml-containing-local-tags-with-ruamel-yaml$
# https://app.soos.io/research/packages/Python/-/pydantic-yaml-parser/

import yaml
import yaml_include
from dumper        import dump
from enum          import Enum
from typing        import List, Dict, Optional, Union, Literal
from pprintpp      import pformat
from pydantic      import BaseModel, Field, ValidationError, ConfigDict
import json
from .common       import *

description = {
    "schema": """
    Configuration file follows YAML syntax.
    UNIX environmenent variables can be accessed anywhere in the YAML config through the `${<env-var>}` syntax.
    The use of UNIX environment variables permits the portability of kgsteward config file, as local path are supplied through variables.
    For security reason, kgsteward checks that every used variable exitsts and is not an empty string.
    In addition, to the environment variables available at statup of kgsteward, a certain number of variables are defined at run-time:
    `${}`, `${}` and `${}`.
    Especially useful is `${dataset.name}` that can be used in be used in `dataset.replace` clause to indicate the current "active" context/named graph.
""",
    "server_brand": """String identifying the server brand. One of 'graphdb', 'rdf4j', 'fuseki' """,
    "location" :  """URL of the server. The SPARQL endpoint locations for queries, updates and stores are specific to a server brand.""" ,
    "repository": """The name of the 'repository' (GraphDB/RDF4J naming) or 'dataset' (fuseki) in the triplestore.""",
    "username":      """The name of a user with write-access rights in the triplestore.""",
    "password":      """The password of a user with write-access rights to the triplestore.
It is recommended that the value of this variable is passed trough an environment variable.
By this way the password is not stored explicitely in the config file.
Alternatively `?` can be used and the password will be asked interactively at run time.""",
    "context_base_IRI": """Base IRI to construct the graph context. In doubt, give `http://example.org/context/` a try.""",
    "file_server_port": """Integer, `0` by default, i.e. the file server is turned off.
When set to a positive integer, say `8000`, local files will be exposed through a temporary
HTTP server and loaded from it. Support for different RDF file types and their compressed
version depend on the tripelstore. The benefit is the that RDF data from `file` are processed
with the same protocol as those supplied remotely through `url`. Essentially for GraphDB,
file-size limits are suppressed and compressed formats are supported.
Beware that the used python-based server is potentially insecure
(see [here](https://docs.python.org/3/library/http.server.html) for details).
This should however pose no real treat if used on a personal computer or on a server that is behind a firewall.
""",
    "server_config":  """Filename with the triplestore configuration, possibly a turtle file.
This file can be saved from the UI interface of RDF4J/GraphDB after a first repository was created interactively,
thus permitting to reproduce the repository configuration elsewhere.
This file is used by the `-I` and `-F` options.
Beware that the repository ID could be hard-coded in the config file and should be maintained in sync with `repository`.
""",
    "dataset": "Mandatory key to specify the content of the knowledge graph in the triplestore.",
    "queries": """A list of paths to files with SPARQL queries to be add to the repository user interface.
Each query is first checked for syntactic correctness by being submitted to the SPARQL endpoint,
with a short timeout.
The query result is not iteself checked.
Wildcards `*` can be used.
""",
    "validations": """A list of paths to files contining SPARQL queries used to validate the repsository.
Wildcards `*` can be used.
By convention, a valid result should be empty, i.e. no row is returned.
Failed results should return rows permitting to diagnose the problems.
""",
    "name": """Mandatory name of a dataset record.""",
    "context": """ The IRI of the target context.
If missing, it will be built by concataining `context_base_IRI` and `name`.
""",
"system": """A list of system command.
This is a simple convenience provided by kgsteward, which is not meant to be a replacement
for serious Make-like system as for example git/dvc.
""",
"file": """List of files containing RDF data.
Wildcards `*` can be used.
The strategy used to load these files will depends on if a file server is used (see `file_server_port` option`).
With GraphDB, there might be a maximum file size (200 MB by default (?)) and compressed files may not be supported.
Using a file server, these limitations are overcomed, but see the security warning described above.
""",
"url": """List of url from which to load RDF data""",
"frozen": "Frozen record, can only be updated explicitely with the `-d <name>` option. The option `-C` has no effect",
"zenodo": """Do not use!
Fetch turtle files from zenodo.
This is a completely ad hoc command developped for ENPKG, that will be suppressed sooner or later.
""",
"update": """List of files containing SPARQL update commands.
Wildcards are not recommended here, as the order of the SPARQL uspdates possibly matters!
""",
#"source": """Path to another kgsteward YAML file from which the graphs list of record will be extracted
#and inserted in the current graphs list.
#""",
    "parent": """A list of dataset names to declare dependency between dataset records.
Updating the parent datset will provoke the update of its children, unless it is frozen.
""",
    "stamp": """List of file paths or URLs to which last modification dates will used.
The file contents are ignored.
Wildcards `*` can be used.
""",
    "replace": """Dictionary to perform string substitution in SPARQL queries from `update` list.
Of uttermost interest is the `${TARGET_GRAPH_CONTEXT}` which permit to restrict updates to the current context.
""",
    "prefixes": """
        A list of Turtle files from which prefix definitions can be obtained.
        This list will used to update the namespace definitions in GraphDB and RDF4J.
        Otherwise it is ignored
    """,
    "sparql_file_loader" : """Files are loaded using the SPARQL update statement: "LOAD <file://<file-path> INTO...". This strategy is likely to failed for large files, or worst silently truncate them.""",
    "store_file_loader"  : """Files are loaded using the graph store protocol. This strategy is likely to failed for large files, or worst silently truncate them. """,
    "http_file_server"   : """Files are exposed through a temproray HTTP server. This is the recommended method with GraphDB.""",
    "riot_chunk_store"   : """Files are parsed through riot (part of JENA distribution), and submitted by chunks using graph store protocol. This is the recommended method with Fuseki. """,
    "sparql_url_loader"  : """URL are loaded using the SPARQL update statement: "LOAD <url> INTO...". This strategy could fail for large files, or worst silently truncate them. """,
    "curl_riot_chunk_store" : """URL are downloaded using curl to a temporary file, which is then loaded with `riot_chunk_store` method.""",
}

description["location_graphdb"] = description["location"] + " GraphDB has location 'http://localhost:7200' by default"
description["location_fuseki"]  = description["location"] + " Fuseki has location 'http://localhost:3030' by default"
description["location_rdf4j"]   = description["location"] + " RDF4J has location 'http://localhost:8080' by default"

def describe( term ):
    if term in description:
        return description[ term ].replace( "\n", " " ).strip()
    else:
        stop_error( "No description for: " + term )

class GraphDBConf( BaseModel ):
    model_config = ConfigDict( extra='allow' )
    brand             : Literal[ "graphdb" ] = Field( title = "GraphDB brand", description = describe( "server_brand" ))
    location          : str = Field( title = "Server URL", description = describe( "location_graphdb" ))
    server_config     : str = Field( title = "Server config file", description = describe( "server_config" ))
    file_server_port  : Optional[ int ] = Field( None, title = "file_server_port", description = describe( "file_server_port" ))
    username          : Optional[ str ] = Field( None, title = "Username", description = describe( "username" ))
    password          : Optional[ str ] = Field( None, title = "Password", description = describe( "password" ))
    prefixes          : Optional[ list[str]]  = Field( None, title = "GraphDB namespace", description = describe( "prefixes" ))
    repository        : str= Field( pattern = r"^\w{1,32}$", title = "Repository ID", description = describe( "repository" ))

class FusekiConf( BaseModel ):
    model_config = ConfigDict( extra='allow' )
    brand             : Literal[ "fuseki" ] = Field( title = "Fuseki brand", description = describe( "server_brand" ))
    location          : str = Field( title = "Server URL", description = describe( "location_fuseki" ))
    repository        : str= Field( pattern = r"^\w{1,32}$", title = "Repository ID", description = describe( "repository" ))
    file_server_port  : Optional[ int ]  = Field( 0, title = "file_server_port", description = describe( "file_server_port" ))

class RDF4JConf( BaseModel ):
    model_config = ConfigDict( extra='allow' )
    brand             : Literal[ "rdf4j" ] = Field( title = "RDF4J brand", description = describe(  "server_brand" ))
    location          : str = Field( title = "Server URL", description = describe( "location_rdf4j" ))
    repository        : str= Field( pattern = r"^\w{1,32}$", title = "Repository ID", description = describe( "repository" ))
    file_server_port  : Optional[ int ]  = Field( 0, title = "file_server_port", description = describe( "file_server_port" ))

class DatasetConf( BaseModel ):
    name     : str = Field( pattern = r"^[a-zA-Z]\w{0,31}$", title = "Short name of a dataset reccord", description = describe( "name" ))
    context  : Optional[ str ]        = Field( None,  title = "Full IRI of a context/named graph", description = describe(  "context" ))
    parent   : Optional[ List[ str ]] = Field( None,  title = "Parent(s) of a dataset record", description = describe(  "parent" ))
    frozen   : Optional[ bool ]       = Field( default = False,  title = "Frozen dataset record", description = describe( "frozen") )
    system   : Optional[ List[ str ]] = Field( None,  title = "UNIX system command(s)", description = describe(  "system" ))
    file     : Optional[ list[ str ]] = Field( None,  title = "Load RDF from file(s)", description = describe(  "file" ))
    url      : Optional[ list[ str ]] = Field( None,  title = "Load RDF from URL(s)", description = describe(  "url" ))
    stamp    : Optional[ list[ str ]] = Field( None,  title = "Stamp file(s)", description = describe(  "stamp" ))
    replace  : Optional[ dict [ str, str ]] = Field( None, title = "String subtitution in SPARQL update(s)", description = describe(  "replace" ))
    update   : Optional[ list[ str ]] = Field( None,  title = "SPARQL update file(s)", description = describe(  "update" ))
    zenodo   : Optional[ list[ int ]] = Field( None,  title = "Ignore me", description = describe(  "zenodo" ))

class SparqlFileLoader( BaseModel ):
    method : Literal[ "sparql_load" ] = Field( title = "sparql file loader", description = describe( "sparql_file_loader" ) )
class StoreFileLoader( BaseModel ):
    method : Literal[ "file_store" ] = Field( title = "store file loader", description = describe( "store_file_loader" ) )
class HttpServerFileLoader( BaseModel ):
    method : Literal[ "http_server" ] = Field( title = "HTTP file server", description = describe( "http_file_server" ))
    port : Optional[ int ] = Field( 8000, title = "file_server_port", description = describe( "file_server_port" ))
class RiotChunkStoreFileLoader( BaseModel ):
    method : Literal[ "riot_chunk_store" ] = Field( title = "riot/store file loader", description = "riot_chunk_store" )
    size : Optional[ int ] = Field( 100_000_000, title = "chunk size", description = "chunk size" )

class SparqlUrlLoader( BaseModel ):
    method : Literal[ "sparql_load" ] = Field( title = "direct url loader", description = describe( "sparql_url_loader" ))
class CurlRiotChunkStoreUrlLoader( BaseModel ):
    method : Literal[ "curl_riot_chunk_store" ] = Field( title = "curl/riot/store URL loader", description = describe( "curl_riot_chunk_store" ))
    tmp_dir : Optional[ str ] = Field( "/tmp", title = "temporary directory", description = "temporary directory" )
    size : Optional[ int ] = Field( 100_000_000, title = "chunk stize", description = "chunk size" )

class KGStewardConf( BaseModel ):
    model_config = ConfigDict( extra='allow' )
    version           : Literal[ "kgsteward_yaml_2" ] = Field( title = "YAML syntax version", description = "This mandatory fixed value determines the admissible YAML syntax" )
    server            : Union[ GraphDBConf, RDF4JConf, FusekiConf ] = Field( discriminator = 'brand' )
    file_loader       : Union[ SparqlFileLoader, StoreFileLoader, HttpServerFileLoader, RiotChunkStoreFileLoader ]
    url_loader        : Union[ SparqlUrlLoader, CurlRiotChunkStoreUrlLoader ]
    dataset           : list[ DatasetConf ] = Field( title = "Knowledge Graph content", description = describe( "dataset" ))
    context_base_IRI  : str = Field( title = "context base IRI", description = describe( "context_base_IRI" ) )
    queries           : Optional[ list[ str ]]  = Field( None, title = "GraphDB queries", description = describe( "queries" ))
    validations       : Optional[ list[ str ]]  = Field( None, title = "Validation queries", description = describe( "validations" ))

def parse_yaml_conf( path : str ):
    """Parsing kgsteward YAML config file(s) is a three step process:
    1. Parse YAML file(s) and execute !include directive
    2. Test YAML version
    3. Content-aware validation through pydantic
    """
    dir_yaml, filename = os.path.split( os.path.abspath( path ))
    file = open( path )
    yaml.add_constructor( "!include", yaml_include.Constructor( base_dir = dir_yaml ))
    try:
        with open( path ) as f:
            data = yaml.full_load(f)
    except Exception as e:
        stop_error( "Something goes wrong with YAML parsing: " + repr( e ))
    if "version" not in data:
        stop_error( 'Key "version" not found in YAML file: ' + path )
    try:
        config = KGStewardConf( **data ).model_dump( exclude_none = True )
    except ValidationError as e:
        stop_error( "YAML syntax error in file " + path + ":\n" + pformat( e.errors() ))

    config["kgsteward_yaml_directory"] = dir_yaml
    config["kgsteward_yaml_filename"]  = filename
    if "username" in config["server"] and not "password" in config["server"] :
        config["server"]["password"] = getpass.getpass( prompt = "Enter password : " )

    dataset = list()
    seen   = []
    for item in config["dataset"] :
        if item["name"] in seen:
            stop_error( "Duplicated name: " +  item["name"] )
        if "parent" in item:
            for parent in item["parent"]:
                if parent == "*":
                    item["parent"] = seen.copy()
                    break
                else:
                    if not parent in seen:
                        stop_error( "Parent not previously defined: " + parent )
        seen.append( item["name"] )
        if "context" not in item:
            item["context"] = str( config["context_base_IRI"] ) + str( item["name"] )
        dataset.append( item )
    config["dataset"] = dataset
    return config

def save_json_schema( path ):
    schema = KGStewardConf.model_json_schema()
    schema["description"] = describe( "schema" )
    with open( path, "w" ) as f:
        f.write( dumps( schema, indent=2 ))



# --------------------------------------------------------- #
# An RDF4J driver
# --------------------------------------------------------- #

from dumper import dump
from .common import *

from .generic import GenericClient

class RFD4JClient( GenericClient ):

    def __init__( self, server_url, username, password, repository ):
        super().__init__(  
            server_url + "/repositories/" + repository,
            server_url + "/repositories/" + repository + "/statements",
            server_url + "/repositories/" + repository + "/rdf-graphs/service"
        )
        self.server_url    = server_url
        self.username      = username
        self.password      = password
        self.repository    = repository
        self.headers       = {} # to be updated below
        print_break()
        print_task( "Contacting server" )
        try:
            http_call({
                'method' : 'GET',
                'url'    : self.server_url
            }, [ 200 ])
        except:
            stop_error( "Cannot contact server at location: " + self.server_url )
#        try:
#            http_call({
#                'url'    : self.endpoint_query
#                'method' : 'GET',
#        except:
#            }, [ 200 ])
#            stop_error( "Server is running, but repository does not exist: " + self.repository )

    def rewrite_repository( self, rdf4j_config_filename ):
        # FIXME: this does not work!
        try: # attempt to erase the repo and its content
            # curl -X DELETE http://localhost:8080/rdf4j-server/repositories/TEST
            http_call({ # slower alternative self.sparql_update( "DROP SILENT ALL" )
                'method' : 'DELETE',
                'url'    : self.server_url + '/repositories/' + self.repository,
                'headers' : self.headers
            }, [ 204, 404 ] ) # 204: cleared; 404 unknown
        except: # case repo does not exist yet 
            pass
        # "curl -H 'content-type: text/turtle' --upload-file common/data/config/JLW_Native_Lucene.config.ttl http://localhost:8080/rdf4j-server/repositories/JLW_Native_Lucene"
        http_call({
            'method'  : 'PUT',
            'url'     : self.server_url + '/repositories/' + self.repository,
            'headers' : { **self.headers, "content-type": "text/turtle" },
            'data'    : open( rdf4j_config_filename , 'rb' )
        }, [ 204, 409 ] ) # 204: created; 409 repository already exists (and don't care)

    def sparql_query( self, sparql, status_code_ok = [ 200 ], echo = True, timeout = None ):
        if echo :
            print_strip( sparql.replace( "\t", "    " ), color = "green" )
        headers = {
            'Accept' : 'application/json', 
            'Content-Type': 'application/x-www-form-urlencoded' 
        }
        params = { 'query' : sparql }
        if timeout is not None :
            params["timeout"] = timeout
            status_code_ok.append( 503 )
            status_code_ok.append( 500 ) # is returned by GraphDB on timeout of SPARQL queries with a SERVICE clause 
                                         # FIXME: verify if rdf4j behave the same
        r = http_call(
            {
                'method'  : 'POST',  # allows for big query string
                'url'     : self.endpoint_query,
                'headers' : { **self.headers, **headers },
                'params'  : params,
            },
            status_code_ok
        )
        if timeout is not None:
            if r.status_code == 503 :
                time.sleep( 1 )
                print_warn( "query timed out" )
                return None
            elif r.status_code == 500 : 
                time.sleep( 1 )
                print_warn( "unknown error, maybe timeout" )
                return None
        return r

    def sparql_update( self, sparql,status_code_ok = [ 204 ], echo = True ):
        if echo :
            print_strip( sparql.replace( "\t", "    " ), color = "green" )
        http_call(
            {
                'method'  : 'POST',
                'url'     : self.endpoint_update,
                'headers' : self.headers,
                'params'  : { 'update': sparql },
            }, 
            status_code_ok,
            echo
        )

    def load_from_file( 
        self,
        file,
        context, 
        headers = {},
        echo = True 
    ):
        super().load_from_file( file, context, { **self.headers, **headers }, echo )
    
    def list_context( self, echo = True ) :
        r = http_call({
            'method'  : 'GET',
            'url'     : self.server_url + "/repositories/" + self.repository + "/contexts",
            'headers' : { **self.headers, 'Accept': 'application/json' },
        }, [ 200 ], echo )
        contexts = set()
        for rec in r.json()["results"]["bindings"] : 
            contexts.add( rec["contextID"]["value"] )
        return contexts

    def drop_context( self, context, echo = True ):
        self.sparql_update( f"DROP GRAPH <{context}>", echo = echo )

    def rewrite_prefixes( self, echo = True ):
        return
        r = http_call({
            'method'  : 'DELETE',
            'url'     : self.server_url + "/repositories/" + self.repository + "/namespaces",
            'headers' : self.headers,
        }, [204], echo )

    def set_prefix( self, short, long, echo = True ):
        return
        r = http_call({
            'method'  : 'PUT',
            'url'     : self.server_url + "/repositories/" + self.repository + "/namespaces/" + short,
            'headers' : self.headers.copy().update({ 'Accept', 'text/plain' }),
            'data'    : long
        }, [204], echo )





# --------------------------------------------------------- #
# Some REST services were not documented in the help pages
# of GraphDB. The syntax of these services can be deduced
# from the JS code of the graphdb-workbench at GitHub
#
#    https://github.com/Ontotext-AD/graphdb-workbench
#
# in directory /src/js/angular/rest, or better in
#
# https://rdf4j.org/documentation/reference/rest-api/
#
# If one mess up with user/password, GraphDB may end up in
# a state where connection become impossible. The only
# way out is to erase the graphdb.home directory
# (Mac in ~/Library/Application\ Support/GraphDB)
# and restart GraphDB. All data will be lost!
#
# Given the frequent updates of GraphDB, some of the 
# above comments might already be deprecated.
# --------------------------------------------------------- #

import time

from dumper import dump
from .common import *
import urllib.parse

from .generic import GenericClient

class GraphDBClient( GenericClient ):

    def __init__( self, graphdb_url, username, password, repository_id ):
        # TODO: ping server first and produce a dcent error message!
        super().__init__( 
            graphdb_url + "/repositories/" + repository_id,
            graphdb_url + "/repositories/" + repository_id + "/statements",
            graphdb_url + "/repositories/" + repository_id + "/rdf-graphs/service"
        )
        self.graphdb_url          = graphdb_url
        self.username             = username
        self.password             = password
        self.repository_id        = repository_id
        self.headers              = {} # to be updated below
        print_break()
        print_task( "contacting server" )
        if self.username is not None :
            r = http_call({
                'method'  : 'POST',
                'url'     : self.graphdb_url + "/rest/login/" + self.username,
                'headers' : {
                    "X-GraphDB-Repository" : self.repository_id,
                    "X-GraphDB-Password"   : self.password
                }
            })
            if 'Authorization' in r.headers:
                self.headers = { 'Authorization': r.headers[ 'Authorization' ] }
            else:
                raise RuntimeError(
                    f"Authentication to GraphDB server failed: {self.graphdb_url}"
                )

    def rewrite_repository( self, graphdb_config_filename ) :
        http_call({
            'method' : 'DELETE',
            'url'    : self.graphdb_url + '/rest/repositories/' + self.repository_id,
            'headers' : self.headers
        })
        http_call({
            'method'  : 'POST',
            'url'     : self.graphdb_url + '/rest/repositories',
            'headers' : self.headers,
            'files'   : { 'config' : open( graphdb_config_filename , 'rb' )}
        }, [ 201 ] )
        http_call({
            'method'  : 'POST',
            'url'     : self.graphdb_url + '/rest/autocomplete/enabled',
            'headers' : { **self.headers, "X-GraphDB-Repository": self.repository_id },
            'params'  : { 'enabled' : 'true' }
        })
        http_call({
            'method'  : 'POST',
            'url'     : self.graphdb_url + "/rest/security",
            'headers' : { **self.headers, "X-GraphDB-Repository": self.repository_id },
            'json'    : "true"
        })

    def free_access( self ) :
        # http_call({
        #     'method'  : 'GET',
        #     'url'     : self.graphdb_url + "/rest/class-hierarchy?doReload=true&graphURI=",
        #     'headers' : { **self.headers, "X-GraphDB-Repository": self.repository_id },
        # })
        http_call({
            'method'  : 'POST',
            'url'     : self.graphdb_url + "/rest/security/free-access",
            'headers' : self.headers,
            "json"    : {
                "enabled" : "true",
                "authorities" : [ "READ_REPO_" + self.repository_id ],
                "appSettings" : {
                    "DEFAULT_INFERENCE"        : "true",
                    "DEFAULT_VIS_GRAPH_SCHEMA" : "true",
                    "DEFAULT_SAMEAS"           : "true",
                    "IGNORE_SHARED_QUERIES"    : "false",
                    "EXECUTE_COUNT"            : "true"
                }
            }
        })

    def sparql_query( self, sparql, status_code_ok = [ 200 ], echo = True, timeout = None ):
        if echo :
            print_strip( sparql.replace( "\t", "    " ), color = "green" )
        headers = {
            'Accept' : 'application/json', 
            'Content-Type': 'application/x-www-form-urlencoded' 
        }
        params = { 'query' : sparql }
        if timeout is not None :
            params["timeout"] = timeout
            status_code_ok.append( 503 )
            status_code_ok.append( 500 ) # is returned by GraphDB on timeout of SPARQL queries with a SERVICE clause ?!?
        r = http_call(
            {
                'method'  : 'POST',  # allows for big query string
                'url'     : self.endpoint_query,
                'headers' : { **self.headers, **headers },
                'params'  : params,
            },
            status_code_ok
        )
        if timeout is not None:
            if r.status_code == 503 :
                time.sleep( 1 )
                print_warn( "query timed out" )
                return None
            elif r.status_code == 500 : 
                time.sleep( 1 )
                print_warn( "unknown error, maybe timeout" )
                return None
        return r
    
    def sparql_update( self, sparql, status_code_ok = [ 204 ], echo = True ):
        if echo :
            print_strip( sparql.replace( "\t", "    " ), color = "green" )
        http_call(
            {
                'method'  : 'POST',
                'url'     : self.endpoint_update,
                'headers' : self.headers,
                'params'  : { 'update': sparql },
            }, 
            status_code_ok,
            echo
        )

    def load_from_file( 
        self,
        file,
        context, 
        headers = {},
        echo = True 
    ):
        super().load_from_file( file, context, { **self.headers, **headers }, echo )

    def load_from_file_using_riot( self, file, context, echo = True ):
        super().load_from_file_using_riot( file, context, headers = { **self.headers }, echo = echo )
       
    def list_context( self, echo = True ) :
        r = http_call({
            'method'  : 'GET',
            'url'     : self.graphdb_url + "/repositories/" + self.repository_id + "/contexts",
            'headers' : { **self.headers, 'Accept': 'application/json' }
        }, [ 200 ], echo )
        contexts = set()
        for rec in r.json()["results"]["bindings"] : 
            contexts.add( rec["contextID"]["value"] )
        return contexts

    def drop_context( self, context, echo = True ):
        self.sparql_update( f"DROP GRAPH <{context}>", echo = echo )

    def graphdb_call( self, request_args, status_code_ok = [ 200 ], echo = True ) :
        request_args['url'] = self.graphdb_url + str( request_args['url'] )
        if 'headers' in request_args :
            request_args['headers'] ={ **self.headers, **request_args['headers'] }
        else :
            request_args['headers'] = self.headers
        return http_call( request_args, status_code_ok, echo )

    def rewrite_prefixes( self, echo = True ):
        r = http_call({
            'method'  : 'DELETE',
            'url'     : self.graphdb_url + "/repositories/" + self.repository_id + "/namespaces",
            'headers' : self.headers,
        }, [204], echo )

    def set_prefix( self, short, long, echo = True ):
        r = http_call({
            'method'  : 'PUT',
            'url'     : self.graphdb_url + "/repositories/" + self.repository_id + "/namespaces/" + short,
            'headers' : { **self.headers, 'Accept': 'text/plain' },
            'data'    : long
        }, [204], echo )





import pytest
from testcontainers.core.container import DockerContainer
from testcontainers.core.waiting_utils import wait_for_logs

from . import run_cmd, env

# Stop and delete all testcontainers: docker stop $(docker ps -a -q) && docker rm $(docker ps -a -q)
# NOTE: in case issue in rootless docker: https://github.com/testcontainers/testcontainers-python/issues/537
# TESTCONTAINERS_DOCKER_SOCKET_OVERRIDE=/run/user/$(id -u)/docker.sock uv run pytest -s

TRIPLESTORE_IMAGE = 'ontotext/graphdb:10.8.5' # latest 10 release with no license

env["GRAPHDB_USERNAME"] = "admin"
env["GRAPHDB_PASSWORD"] = "root"

@pytest.fixture( scope="module" )
def triplestore():
    """Start GraphDB container as a fixture."""
    container = DockerContainer(TRIPLESTORE_IMAGE)
    container.with_exposed_ports(7200).with_bind_ports(7200, 7200)
    container.with_env("JAVA_OPTS", "-Xms1g -Xmx4g")
    container.start()
    delay = wait_for_logs(container, "Started GraphDB")
    # host = container.get_container_host_ip()
    # port = container.get_exposed_port(7200)
    # base_url = f"http://{host}:{port}"
    base_url = f"http://{container.get_container_host_ip()}:{container.get_exposed_port(7200)}"

    print(f"GraphDB started in {delay:.0f}s at {base_url}")
    # print(container.get_logs())
    yield base_url

cmd_base = [
    "kgsteward doc/first_steps/graphdb.yaml -I -v", # Initialize repository
    "kgsteward doc/first_steps/graphdb.yaml -C -v", # Complete (populate) repository
    "kgsteward doc/first_steps/graphdb.yaml -V -v", # Validate repository
    "kgsteward doc/first_steps/graphdb.yaml -Q -v", # validate Queries 
    "mkdir -p tmp/first_steps",
    "rm -f tmp/first_steps/*.tsv",
    "kgsteward doc/first_steps/graphdb.yaml -x tmp/first_steps -v", # Serialize query results for testing
    "diff -r doc/first_steps/ref tmp/first_steps"
]

cmd_graphdb = [
    "kgsteward doc/first_steps/graphdb.yaml --graphdb_upload_queries -v",
    "kgsteward doc/first_steps/graphdb.yaml --graphdb_upload_prefixes -v",
    "kgsteward doc/first_steps/graphdb.yaml --graphdb_free_access -v",
]

@pytest.mark.parametrize( "cmd", cmd_base + cmd_graphdb )
def test_run_cmd_graphdb( triplestore , cmd):
    print( "##############################################################################" )
    print( "### " + cmd )
    print( "##############################################################################" )
    res = run_cmd( cmd.split( " " ), env )
    print(res.stdout)
    print(res.stderr)
    assert res.returncode == 0



import os
import glob
import requests
import shutil
import time
import re
import dumper
import sys
import gzip
import bz2
import lzma

from termcolor import colored

RE_CATCH_ENV_VAR     = re.compile( r"\$\{([^\}]+)\}" )
RE_CATCH_BEGIN_SPACE = re.compile( r"^(\s*)" )
RE_CATCH_END_SPACE   = re.compile( r"(\s*)$" )
RE_CATCH_TAB         = re.compile( r"(\t)" )

def replace_env_var( txt ) :
    """ A helper sub with no magic """
    m = RE_CATCH_ENV_VAR.search( txt )
    if m:
        env_var = m.group( 1 )
        if env_var in os.environ:
            return replace_env_var( txt.replace( "${" + env_var + "}", os.getenv( env_var ))) # recursion
        else:
            stop_error(f"Environment variable not set: {env_var}")
    else:
        return txt

def update_path( path, default_dir, fatal = True  ):
    """Expand path environment variables"""
    dir, fn = os.path.split( os.path.normpath( replace_env_var( path ))) # handles ".." correctly
    if not dir.startswith( "/" ):
        dir = default_dir + "/" + dir
    filename = dir + "/" + fn
    if not os.path.isfile( filename ):
        stop_error( "File does not exists: " + filename )
    return filename

def expand_path( path, default_dir, fatal = True ):
    """Expand path on wildcard and environment variables"""
    dir, filename = os.path.split( os.path.normpath( replace_env_var( path ))) # handles ".." correctly
    res = []
    if not dir.startswith( "/" ):
        dir = default_dir + "/" + dir
    if "*" in dir or  "*" in filename:
        paths = glob.glob( dir + "/" + filename )
        if paths:
            for p in sorted( paths ):
                res.append( os.path.split( p ))
        else:
            if fatal:
                stop_error( "Not a single file found: " + dir + "/" + filename )
            else:
                print_warn( "Not a single file found: " + dir + "/" + filename )
    else:
        if os.path.isfile( dir + "/" + filename ):
            res.append([ dir, filename ])
        else:
            if fatal:
                stop_error( "File does not exists: " + dir + "/" + filename )
            else:
                print_warn( "File does not exists: " + dir + "/" + filename )
    return res

def http_call( request_args, status_code = [ 200 ], echo = True ):
    """A simple wrapper arround requests.request() which main purpose is to
    (1) validate status code
	(1) print elapsed time
    (2) print request parameters in case of an unexpected status code."""
    if echo :
        report( request_args['method'], request_args['url'] )
    start_time = time.time()
    r = requests.request( **request_args )
    end_time = time.time()
    if r.status_code not in status_code :
        dumper.dump( request_args )
        print_warn( "Status code = " + str( r.status_code ))
        print_warn( r.text if not r.text is None else '' )
        raise Exception( "HTTP request failed!" )
    if echo :
        report( "elapsed time", ( "%.3f" % ( end_time - start_time )))
    return r

def get_head_info( url, echo = True ):
    r = http_call({ 'method': 'HEAD', 'url': url }, echo = echo )
    str = url
    for key in sorted( r.headers ):
         if key.lower() in [ "last-modified",  "content-length" ]:
            str += " " + key + " " + r.headers[ key ]
    return str

def download_file( url, filename ):
    with requests.get(url, stream=True) as r:
        with open( filename, 'wb') as f:
            shutil.copyfileobj(r.raw, f)

def any_open( filename, mode = "rb"):
    if filename.endswith(".gz"):
        return gzip.open( filename, mode )
    elif filename.endswith(".bz2"):
        return bz2.open( filename, mode )
    elif filename.endswith(".xz"):
        return lzma.open( filename, mode )
    else:
        return open( filename, mode )

# FROM https://graphdb.ontotext.com/documentation/10.7/rdf-formats.html
def guess_mime_type( filename ):
    if re.search( r"\.ttl(|\.gz|\.bz2|\.xz)$", filename ):
       return "text/turtle"
    elif re.search( r"\.ttls(|\.gz|\.bz2|\.xz)$", filename ):
       return "text/x-turtlestar"
    elif re.search( r"\.(trig)(|\.gz|\.bz2|\.xz)$", filename ):
        return "application/trig"
    elif re.search( r"\.(trigs)(|\.gz|\.bz2|\.xz)$", filename ):
        return "application/x-trigstar"
    elif re.search( r"\.(n3)(|\.gz|\.bz2|\.xz)$", filename ):
        return "text/rdf+n3"
    elif re.search( r"\.(nt)(|\.gz|\.bz2|\.xz)$", filename ):
        return "application/n-triples"
    elif re.search( r"\.(nq)(|\.gz|\.bz2|\.xz)$", filename ):
        return "application/n-quads"
    elif re.search( r"\.(jsonld)(|\.gz|\.bz2|\.xz)$", filename ):
        return "application/ld+json"
    elif re.search( r"\.(ndjsonld|jsonl|ndjson)(|\.gz|\.bz2|\.xz)$", filename ):
        return "application/x-ld+ndjson"
    elif re.search( r"\.(rj)(|\.gz|\.bz2|\.xz)$", filename ):
        return "application/rdf+json"
    elif re.search( r"\.(rdf|rdfs|owl|xml)(|\.gz|\.bz2|\.xz)$", filename ):
        return "application/rdf+xml"
    elif re.search( r"\.(trix)(|\.gz|\.bz2|\.xz)$", filename ):
        return "application/trix"
    elif re.search( r"\.(brf)(|\.gz|\.bz2|\.xz)$", filename ):
        return "application/x-binary-rdf"
    else:
        stop_error( "cannot guess RDF mime-type from filename: " + filename )


capture_block  = re.compile( r"(.+);\s*(\#[^\n]+\n)*\s*(LOAD|CLEAR|DROP|CREATE|ADD|MOVE|COPY|WITH|DELETE|INSERT)(.+)$", re.IGNORECASE | re.MULTILINE | re.DOTALL )
capture_prefix = re.compile( r"(PREFIX\s+\S+\s+\S+)\s*", re.IGNORECASE | re.MULTILINE | re.DOTALL )

def split_sparql_update( sparql ): # FIXME: handle comments 
    statement = []
    prefix    = []
    m = capture_prefix.search( sparql )
    while m:
        prefix.append( m[1] )
        sparql = sparql.replace( m[1], '' )
        m = capture_prefix.search( sparql )        
    m = capture_block.search( sparql )
    while m:
        statement.append( m[3] + m[4] )
        sparql = m[1] 
        m = capture_block.search( sparql )
    statement.append( sparql )
    for i in range( len( statement )):
        statement[i] = re.sub( r"^\n+", '', statement[i] )
        statement[i] = re.sub( r"\n+$", '', statement[i] )
        statement[i] = "\n".join( prefix ) + "\n" + statement[i]
    return reversed( statement )

def sparql_result_to_table( r ):
    """Reformat SPARQL results as JSON into a table.
       Standardize data representation as much as possible, to counteract server fantasy.
       Not optimized for speed, however"""
    try:
        j = r.json()
        tbl = []
        header = j["head"]["vars"]
        for rec in j["results"]["bindings"] :
            row = []
            for h in header:
                if h not in rec:
                    row.append( "" )
                elif rec[h]["type"] == 'uri':
                    row.append( "<" + rec[h]["value"] + ">" )
                else: # FIXME: standardize data representation here
                    row.append( rec[h]["value"] )
            tbl.append( row )
        return header, tbl
    except Exception as e :
        dumper.dump( j )
        dumper.dump( e )
        stop_error( "Parsing went wrong!" )

def print_break():
    print( '# ----------------------------------------------------------------------------')

def print_strip( txt, color = 'black' ):
    """Print after removing leading/trailing spaces"""
    txt = RE_CATCH_TAB.sub( '    ', RE_CATCH_END_SPACE.sub( '', RE_CATCH_BEGIN_SPACE.sub( '', txt )))
    if color :
        print( colored( txt, color ), flush = True )
    else :
        print( txt , flush = True )

def report( key, value, color = "black" ) :
    print( '# ' + colored( f"{key:>12} : {value}" , color ), flush = True )

def print_task( txt ):
    report( 'TASK', txt, color = "black" )

def print_warn( txt ):
    report( 'warning', txt, color = "red" )

def stop_error( txt ):
    report( 'ERROR', txt, color = "red" )
    sys.exit( 1 )



"""Main kgsteward module."""

import argparse
import os
import sys
import re
import hashlib
import requests  # https://docs.python-requests.org/en/master/user/quickstart/
import subprocess
import urllib

from   dumper    import dump # get ready to help debugging
from   termcolor import colored

from .common     import *
from .yamlconfig import parse_yaml_conf
from .graphdb    import GraphDBClient
from .fuseki     import FusekiClient
from .rdf4j      import RFD4JClient
# from .oxigraph   import OxigraphClient  # in preparation
# from .virtuoso   import VirtuosoClient  # in preparation
from .fileserver import LocalFileServer

name2context = {} # global helper dict
context2name = {} # same

# ---------------------------------------------------------#
# Command-line options
# ---------------------------------------------------------- #

def get_user_input():
    """Generate argparse CLI and return user input."""

    parser = argparse.ArgumentParser(
        description = "A command line tool to manage the content of a RDF store, as specified in a YAML configuration file. "
                      "Documentations, including supported YAML syntax and the source are available from https://github.com/sib-swiss/kgsteward."
    )
    parser.add_argument(
        'yamlfile',
        nargs = 1,
        help  = "Mandatory configuration file in YAML format. "
                "Supported syntax: https://github.com/sib-swiss/kgsteward/blob/main/doc/yaml/kgsteward.schema.md"
    )
    parser.add_argument(
        '-F',
        action = 'store_true',
        help   = "Force rebuild the core of the RDF database. "
                 "Sames as -I -D combined options."
    )
    parser.add_argument(
        '-I',
        action = 'store_true',
        help   = "Create the repository or force rewrite the repository, i.e. erase all existing RDF data."
    )
    parser.add_argument(
        '-D',
        action = 'store_true',
        help   = 'Force reload all RDF data'
    )
    parser.add_argument(
        '-d',
        help = "Force reload RDF data from the supplied comma-separated list "
               "of graph name (no space in the list). It might be combined "
               "with option -C.",
    )
    parser.add_argument(
        '-C',
        action = 'store_true',
        help   = "Load missing/incomplete/outdated dataset to complete the repository."
    )
    parser.add_argument(
        '-U',
        action = 'store_true',
        help   = 'Force update internal checksum without updating the data. '
    )
    parser.add_argument(
        '-V',
        action = 'store_true',
        help   = "Validate the repository content, by running SPARQL queries that supposed to return the problems. "
                 "It might be combined with --timeout."
    )
    parser.add_argument(
        '-Q',
        action = 'store_true',
        help   = "Run all SPARQL queries. It might be combined with --timeout."
    )
    parser.add_argument(
        '-x',
        help   = "Dump all query results in TSV format to dir <X>. "
                 "Sorting of results is enforced, irrespective of the SPARQL queries. "
                 "The results are amended to facilitate the comparison of different server output, e.g. using diff. "
                 "Using this service is useful for debugging, but not meant to retrieve data. " 
                 "It is possibly slow and memory hungry." 
    )
    parser.add_argument(
        '-y',
        help   = "Dump all context data in TSV format to dir <Y>. "                 
                 "Sorting of results is enforced, irrespective of the SPARQL queries. "
                 "The results are amended to facilitate the comparison of different server output, e.g. using diff. "
                 "Using this service is useful for debugging, but not meant to retrieve data. " 
                 "It is possibly slow and memory hungry." 
    )
    parser.add_argument(
        '-v',
        action = 'store_true',
        help    = "Verbose mode: print out SPARQL queries being executed. "
                  "Super useful for debugging SPARQL update (after string replacement)."
    )
    parser.add_argument(
        '--timeout',
        type = int,
        help   = "Timeout delay in seconds. There is no timeout by default. "
                 "Values lower than 15 are not recommended with GraphDB." 
    )
    parser.add_argument(
        '--fuseki_compress_tbd2',
        action = 'store_true',
        help = "Compress fuseki TDB2 indexes, otherwise do nothing. This is executed at first. "
               "Beware of the advantage and disadvantage of TDB vs TDB2 indexes in fuseki."
    )
    parser.add_argument(
        '--graphdb_upload_queries',
        action = 'store_true',
        help = "Rewrite and upload queries in the GraphDB menu. "
               "The queries are global to a GraphDB instance, i.e. they are not attached to a specific repository. "
    )
    parser.add_argument(
        '--graphdb_upload_prefixes', '--rdf4j_upload_prefixes',
        action = 'store_true',
        help = "Rewrite and reload all prefix definitions. "
               "The prefixes are global to a GraphDB/RDF4J instance, i.e. they are not attached to a specific repository. "
    )
    parser.add_argument(
        '--graphdb_free_access',
        action = 'store_true',
        help = "Allow read-only, public free-access to a repository in GraphDB. "
    )
    parser.add_argument(
        '--sib_swiss_editor',
        help = "Document and save all queries and prefix declarations in a single Turtle file, ready to be retrived by the sib-swiss editor (https://github.com/sib-swiss/sparql-editor). "
               "Note that the <SIB_SWISS_EDITOR> file is not uploaded directly to the store. "
    )
    args = parser.parse_args()

    # Further processing of command line arguments
    if args.F :
        args.I = True
        args.D = True

    return args

def get_target( config, name ):
    """ An inefficient helper function """
    for rec in config["dataset"] :
        if rec["name"] == name :
            return rec
    raise RuntimeError( "Target name not found: " + name )

def get_sha256( config, name, echo = True ) :
    """ Compute checksums of dataset record """
    target = get_target( config, name )
    context = name2context[ name ] # get_context( config, name )
    os.environ["TARGET_GRAPH_CONTEXT"] = context
    os.environ["kgsteward_dataset_name"]    = name
    os.environ["kgsteward_dataset_context"] = context
    sha256 = hashlib.sha256()
    if "system" in target :
        for cmd in target["system"] :
            sha256.update( cmd.encode( 'utf-8' ))
    if "url" in target :
        for url in target["url"] :
            path = replace_env_var( url )
            sha256.update( path.encode('utf-8') )
            if re.search( r"https?:", path ) : # do not run HEAD on ftp server FIXME: implement something better
               info = get_head_info( path, echo = echo ) # as a side effect: verify is the server is responding
               sha256.update( info.encode('utf-8') )
    if "stamp" in target :
        for link in target["stamp"] :
            path = replace_env_var( link )
            sha256.update( path.encode('utf-8') )
            if( path.startswith( "http" )):
                info = get_head_info( path, echo = echo  ) # as a side effect: verify is the server is responding
                sha256.update( info.encode('utf-8') )
            else:  # assume local file
                for dir, fn in expand_path( path, config["kgsteward_yaml_directory"], fatal = False ):
                    filename = dir + "/" + fn
                    with open( replace_env_var( filename ), "rb") as f :
                        for chunk in iter( lambda: f.read(4096), b"") :
                            sha256.update( chunk )
    if "file" in target :
        for path in target["file"] :
            for dir, filename in expand_path( path, config["kgsteward_yaml_directory"], fatal = False ):
                with open( replace_env_var( dir + "/" + filename ), "rb") as f :
                    for chunk in iter( lambda: f.read(4096), b"") :
                        sha256.update( chunk )
    if "zenodo" in target : # FIXME: remove this!
        for id in target["zenodo"]:
            r = requests.request( 'GET', "https://zenodo.org/api/records/" + str( id ))
            if not r.status_code == 200 :
                raise RuntimeError( 'GET failed: ' + "https://zenodo.org/api/records/" + str( id ))
            info = r.json()
            for record in info["files"] :
                sha256.update( record["checksum"].encode('utf-8'))
    if "replace" in target:
        for key in sorted( target["replace"].keys()):
            sha256.update( key.encode( 'utf-8' ))
            sha256.update( target["replace"][key].encode( 'utf-8' ))
    if "update" in target :
        for path in target["update"] :
            for dir, fn in expand_path( path, config["kgsteward_yaml_directory"] ):
                filename = dir + "/" + fn
                with open( filename ) as f:
                    sparql = f.read()
                sha256.update( sparql.encode('utf-8'))
    return sha256.hexdigest()

def update_config( server, config, echo = True ) :
    """ Compute current status information about the records in repository.
        Take dependency into account."""
    print_break()
    print_task( "Retrieve current status" )
    name2item = {}
    for item in config["dataset"] :
        name = item["name"]
        name2item[name]    = item
        item["count"]      = ""
        item["date"]       = ""
        item["sha256_old"] = ""
        item["status"]     = "FROZEN" if item["frozen"] else "EMPTY"
    r = server.sparql_query( """
PREFIX void: <http://rdfs.org/ns/void#>
PREFIX dct:  <http://purl.org/dc/terms/>
PREFIX ex:   <http://example.org/>
SELECT ?context ?x ( REPLACE( STR( ?y ), "\\\\..+", "" ) AS ?t ) ?sha256
WHERE{
    ?context a void:Dataset   ;
             void:triples  ?x      ;
             dct:modified  ?y      ;
             ex:has_sha256 ?sha256 .
}
""", echo = echo )
    if r is not None:
        try: 
            for rec in r.json()["results"]["bindings"] :
                if rec["context"]["value"] in context2name:
                    name = context2name[rec["context"]["value"]]
                    item = name2item[ name ]
                else:
                    continue
                item["count"]      = rec["x"]["value"]
                item["date"]       = rec["t"]["value"]
                item["sha256_old"] = str( rec["sha256"]["value"] )
                item["sha256_new"] = get_sha256( config, name, echo = echo )
                if item["sha256_old"] == item["sha256_new"]:
                    item["status"] = "ok"
                elif not item["status"] == "FROZEN":
                    item["status"] = "UPDATE"
        except Exception as e: # test before and remove this typ/excep block 
            stop_error( "Failed parsing server response: " + str( e ))
    for item in config["dataset"] :
        if item["status"] == "ok" and "parent" in item :
            for parent in item["parent"] :
                if name2item[parent]["status"] in { "EMPTY", "UPDATE", "PROPAGATE" }:
                    item["status"] = "PROPAGATE"
    return config

def update_dataset_info( server, config, name, echo = True ) :
    print_break()
    context = name2context[ name ]
    sha256 = get_sha256( config, name, echo = echo )
    server.sparql_update( f"""PREFIX void: <http://rdfs.org/ns/void#>
PREFIX dct:  <http://purl.org/dc/terms/>
PREFIX ex:   <http://example.org/>

INSERT {{
    GRAPH <{context}> {{
        <{context}> a void:Dataset ;
            void:triples               ?c   ;
            dct:modified               ?now ;
            ex:has_sha256              "{sha256}" .
    }}
}}
WHERE {{
    GRAPH <{context}> {{
        SELECT
            ( COUNT( * ) AS ?c )
            ( NOW() AS ?now )
        WHERE {{
            ?s ?p ?o
        }}
    }}
}}
""", echo = echo )

def main():
    """Main function of the kgsteward workflow."""

    # --------------------------------------------------------- #
    # Read command line options
    # --------------------------------------------------------- #

    args = get_user_input()

    # --------------------------------------------------------- #
    # Load YAML config and complete it
    # --------------------------------------------------------- #

    print_break()
    print_task( "read YAML config" );
    config = parse_yaml_conf( replace_env_var( args.yamlfile[0] ))
    for item in config["dataset"]:
        name2context[ item["name"] ] = item["context"]
        context2name[ item["context"] ] = item["name"]

    # --------------------------------------------------------- #
    # Initalise connection with the right triplestore
    # --------------------------------------------------------- #

    if "username" in config["server"]:
        username = replace_env_var( config["server"]["username"] )
        password = replace_env_var( config["server"]["password"] )
    else:
        username = None
        password = None

    if config["server"]["brand"] == "graphdb":
        server = GraphDBClient(
            replace_env_var( config["server"]["location"] ),
            username,
            password,
            replace_env_var( config["server"]["repository"] )
        )
    elif config["server"]["brand"] == "rdf4j":
        server = RFD4JClient(
            replace_env_var( config["server"]["location"] ),
            username,
            password,
            replace_env_var( config["server"]["repository"] )
        )
    elif config["server"]["brand"] == "fuseki":
        server = FusekiClient(
            replace_env_var( config["server"]["location"] ),
            replace_env_var( config["server"]["repository"] ),
            replace_env_var( config["server"]["server_config"] )
        )
#            replace_env_var( config["server"]["location"] ),
#            username,
#            replace_env_var( config["server"]["repository"] )
#            password,
#    elif config["server"]["type"] == "oxigraph":
#        server = OxigraphClient(
#            replace_env_var( config["server"]["location"] )
#        )
    else:
        stop_error( "Unknown server brand: " + config["server"]["brand"] )

    for key in config["server"].keys():
        os.environ[ "kgsteward_server_" + str( key )] = str( config["server"][key] )
    os.environ[ "kgsteward_server_endpoint_query"]  = server.get_endpoint_query()
    os.environ[ "kgsteward_server_endpoint_update"] = server.get_endpoint_update()

    # --------------------------------------------------------- #
    # Create a new empty repository or rewrite an existing one
    # --------------------------------------------------------- #

    if args.I :
        if "server_config" in config["server"]:
            config_file = update_path( config["server"]["server_config"], config["kgsteward_yaml_directory"] )
            server.rewrite_repository( config_file )
        else:
             server.rewrite_repository()

    if args.fuseki_compress_tbd2: # FIXME: check that index type is really TDB2 (and not TDB)
        if config["server"]["brand"] == "fuseki": 
            print_break()
            print_task( "Launch TDB2 compression in the. It may delay execution of next statements" )
            server.fuseki_compress_tdb2()
        else:
            print_warn( "Option --fuseki_compress_tbd2 not supported for server brand: " + config["server"]["brand"] )

    # --------------------------------------------------------- #
    # Establish the list of contexts to update
    # --------------------------------------------------------- #

    rdf_graph_all       = set()
    rdf_graph_to_update = set()

    for target in config["dataset"] :
        rdf_graph_all.add( target["name"] )

    if args.D :
        rdf_graph_to_update = rdf_graph_all
    elif args.d : # status not checked here
        for name in args.d.split( "," ) :
            if name in rdf_graph_all :
                rdf_graph_to_update.add( name )
            else :
                raise stop_error( "Invalid name: " + name )
    elif args.C :
        config = update_config( server, config, echo = args.v ) # may takes a while
        for name in rdf_graph_all :
            target = get_target( config, name )
            if target["status"] in { "EMPTY", "UPDATE", "PROPAGATE" } :
                rdf_graph_to_update.add( name )

    # --------------------------------------------------------- #
    # Drop previous data, upload new data in their respective
    # dataset
    # --------------------------------------------------------- #
    
    is_running_in_a_container = False # initial guess
    
    for target in config["dataset"] :

        name = target["name"]
        if not name in rdf_graph_to_update :
            continue

        print_break()
        print_task( "Update dataset record: " + name )
        context = name2context[ name ]
        server.drop_context( context, echo = args.v )

        os.environ["TARGET_GRAPH_CONTEXT"] = context
        os.environ["kgsteward_dataset_name"]    = name
        os.environ["kgsteward_dataset_context"] = context
        replace = {}

        if "system" in target :
            for cmd in target["system"] :
                cmd2 = replace_env_var( cmd )
                print( colored( cmd2, "cyan" ))
                exit_code = os.system( cmd2 )
                if not exit_code == 0 :
                    raise stop_error( 'System cmd failed: ' + cmd2 )

        if "url" in target :
            for u in target["url"] :
                path = replace_env_var( u )
                if config["url_loader"]["method"] == "curl_riot_store":
                    filename = config["url_loader"]["tmp_dir"] + "/" + path.split('/')[-1]
                    cmd = [ "curl", path, "-o", filename ]
                    print( colored( " ".join( cmd ), "cyan" ))
                    subprocess.run( cmd )
                    server.load_from_file_using_riot( filename, context, echo = args.v )
                else: # direct sparql_load
                    server.sparql_update( f"LOAD <{path}> INTO GRAPH <{context}>", echo = args.v )
                    server.sparql_update( f"""PREFIX void: <http://rdfs.org/ns/void#>
INSERT DATA {{
    GRAPH <{context}> {{
        <{context}> void:dataDump <{path}>
    }}
}}""", echo = args.v )
        if "file" in target :
            if config["file_loader"]["method"] == "http_server":
                fs = LocalFileServer( port = config["file_loader"][ "port" ] )
                for path in target["file"] :
                    for dir, fn in expand_path( path, config["kgsteward_yaml_directory"] ):
                        if not os.path.isfile( dir + "/" + fn ):
                            stop_error( "File not found: " + dir + "/" + fn )
                        fs.expose( dir ) # cache already exposed directory
                        if not is_running_in_a_container:
                            try:
                                path = "http://localhost:" + str( config["file_loader"][ "port" ] ) + "/" + fn
                                server.sparql_update( f"LOAD <{path}> INTO GRAPH <{context}>", echo = args.v )
                            except Exception as e:
                                is_running_in_a_container = True # new guess 
                                print_warn( "Maybe the server is running in a container => let's try again!" )
                        if is_running_in_a_container:
                            # see https://stackoverflow.com/questions/68021524/access-localhost-from-docker-container
                            path = "http://host.docker.internal:" + str( config["file_loader"][ "port" ] ) + "/" + fn
                            server.sparql_update( f"LOAD <{path}> INTO GRAPH <{context}>", echo = args.v )
                        filename = dir + "/" + fn
                        server.sparql_update( f"""PREFIX void: <http://rdfs.org/ns/void#>
INSERT DATA {{
    GRAPH <{context}> {{
        <{context}> void:dataDump <file://{filename}>
    }}
}}""", echo = args.v )
                fs.terminate()
            else: # config["file_loader"]["type"] != "http_server"
                for path in target["file"] :
                    for dir, fn in expand_path( path, config["kgsteward_yaml_directory"] ):
                        filename = dir + "/" + fn
                        if config["file_loader"]["method"] == "sparql_load":
                            server.sparql_update( f"LOAD <file://{filename}> INTO GRAPH <{context}>", echo = args.v )
                        elif config["file_loader"]["method"] == "file_store":
                            server.load_from_file( filename, context, echo = args.v )
                        elif config["file_loader"]["method"] == "riot_chunk_store":
                            server.load_from_file_using_riot( filename, context, echo = args.v )
                        else:
                            raise SystemError( "Unexpected file loader method: " + config["file_loader"]["method"] )
                        server.sparql_update( f"""PREFIX void: <http://rdfs.org/ns/void#>
INSERT DATA {{
    GRAPH <{context}> {{
        <{context}> void:dataDump <file://{filename}>
    }}
}}""", echo = args.v)

        if "zenodo" in target :
            for id in target["zenodo"]:
                r = requests.request( 'GET', "https://zenodo.org/api/records/" + str( id ))
                if not r.status_code == 200 :
                    raise RuntimeError( 'GET failed: ' + "https://zenodo.org/api/records/" + str( id ))
                info = r.json()
                for record in info["files"]:
                    path = "https://zenodo.org/records/" + str( id ) + "/files/" + record["key"]
                    server.sparql_update( f"LOAD <{path}> INTO GRAPH <{context}>", echo = args.v )
                    server.sparql_update( f"""PREFIX void: <http://rdfs.org/ns/void#>
INSERT DATA {{
    GRAPH <{context}> {{
        <{context}> void:dataDump <{path}>
    }}
}}""", echo = args.v )

        if "replace" in target:
            for key in target["replace"]:
                replace[key] = replace_env_var( target["replace"][key] )

        if "update" in target :
            for path in target["update"] :
                for dir, fn in expand_path( path, config["kgsteward_yaml_directory"] ):
                    filename = dir + "/" + fn
                    report( "read", filename )
                    with open( filename ) as f:
                        sparql = f.read()
                    for key in sorted( replace.keys()):
                        sparql = sparql.replace( key, replace[ key ])
                    for s in split_sparql_update( sparql): # split on ";" to execute one statement at a time
                        server.sparql_update( s, echo = args.v )

        update_dataset_info( server, config, name, echo = args.v )

    # --------------------------------------------------------- #
    # Force update namespace declarations
    # --------------------------------------------------------- #

    if args.graphdb_upload_prefixes:
        if not config["server"]["brand"] == "graphdb":
            print_warn( "Option --graphdb_upload_prefixes not supported for server brand: " + config["server"]["brand"] )
        else:
            if "prefixes" in config:
                print_task( "rewrite prefixes" )
                server.rewrite_prefixes()
                catch_key_value = re.compile( r"@prefix\s+(\S*):\s+<([^>]+)>" )
                for filename in config["prefixes"] :
                    report( "parse file", filename )
                    file = open( replace_env_var( filename ))
                    for line in file:
                        match = catch_key_value.search( line )
                        if match:
                            server.set_prefix( match.group( 1 ), match.group( 2 ))
            

    # --------------------------------------------------------- #
    # Force update dataset info
    # --------------------------------------------------------- #

    if args.U :
        for target in config["dataset"] :
            update_dataset_info( target )

    # --------------------------------------------------------- #
    # Run all validation tests
    # --------------------------------------------------------- #

    if args.V:
        print_break()
        print_task( "Run SPARQL queries to validate repository content." )
        exit_code = 0
        for path in config["validations"] :
            for dir, fn in expand_path( path, config["kgsteward_yaml_directory"] ):
                filename = dir + "/" + fn
                print_break()
                report( 'query file', filename )
                with open( filename ) as f: sparql = f.read()
                r = server.sparql_query( sparql, echo = args.v, timeout=args.timeout )
                if args.timeout is not None and r is None: # likely timeout
                    report( "Result", "Unknown" )
                    continue
                header, rows = sparql_result_to_table( r )
                if len( rows ) == 0 :
                    report( "Result", "Pass ;-)" )
                else:
                    if not args.v:
                        print_strip( sparql, color = "green" )
                    print( colored( "\t".join( header ), "red" ))
                    for row in rows:
                        print( colored( "\t".join( map( str, row )), "red" ))
                    report( "Result", "Failed ;-(" )
                    exit_code = 1
        print( '----------------------------------------------------------')
        sys.exit( exit_code )

    # --------------------------------------------------------- #
    # Refresh all GraphDB preloaded queries
    # --------------------------------------------------------- #

    if args.Q:
        if "queries" not in config:
            stop_error( "There are no 'queries' key in config! ")
        for path in config["queries"] :
            for dir, fn in expand_path( path, config["kgsteward_yaml_directory"] ):
                filename = dir + "/" + fn
                print_break()
                with open( filename ) as f: sparql = f.read()
                name = re.sub( r'(.*/|)([^/]+)\.\w+$', r'\2', filename )
                sparql = "## " + name + " ##\n" + sparql # to ease debugging from logs
                report( "validate", filename )
                server.validate_sparql_query( sparql, echo = args.v, timeout=args.timeout )

    if args.graphdb_upload_queries:
        if not config["server"]["brand"] == "graphdb":
            print_warn( "Option --graphdb_upload_queries not supported for server brand: " + config["server"]["brand"] )
        else:
            r = server.graphdb_call({ 'url' : '/rest/sparql/saved-queries', 'method' : 'GET' })
            for item in r.json() :
                print_break()
                report( "remove", item['name'] ) # GraphDB builtin queries cannot be deleted
                server.graphdb_call({
                    'url'    : '/rest/sparql/saved-queries',
                    'method' : 'DELETE',
                    'params' : { 'name': item['name']}
                })
            for path in config["queries"] :
                for dir, fn in expand_path( path, config["kgsteward_yaml_directory"] ):
                    filename = dir + "/" + fn
                    report( "read", filename )
                    print_break()
                    with open( filename ) as f: sparql = f.read()
                    name = re.sub( r'(.*/|)([^/]+)\.\w+$', r'\2', filename )
                    sparql = "## " + name + " ##\n" + sparql # to ease debugging from logs
#                    report( "validate", filename )
#                    server.validate_sparql_query( sparql, echo = args.v, timeout=args.timeout )
#                    if not args.graphdb_upload_query:
#                        continue
                    report( "load", filename )
                    server.graphdb_call({
                        'url'    : '/rest/sparql/saved-queries',
                        'method'  : 'POST',
                        'headers' : {
                            'Content-Type': 'application/json', 
                            'Accept-Encoding': 'identity'
                        },
                        'json'    : { 'name': name, 'body': sparql, "shared": "true" }
                    }, [ 201 ] )

    if args.sib_swiss_editor:
        counter = 0 # to keep track of the original input order 
        prefix = {} # to create a non-redundant list
        catch_key_value_ttl = re.compile( r"@prefix\s+(\S*):\s+<([^>]+)>", re.IGNORECASE )
        catch_key_value_rq  = re.compile( r"PREFIX\s+(\S*):\s+<([^>]+)>",  re.IGNORECASE )
        g = Graph()
        SPARQLQUERY   = Namespace( server.get_endpoint_query() + "/.well-known/sparql-examples/" )
        prefixes_iri  = URIRef(    server.get_endpoint_query() + "/.well-known/prefixes" )
        PREFIX        = Namespace( server.get_endpoint_query() + "/.well-known/prefix/" )
        RDF           = Namespace( "http://www.w3.org/1999/02/22-rdf-syntax-ns#" )
        RDFS          = Namespace( "http://www.w3.org/2000/01/rdf-schema#" )
        SCHEMA        = Namespace( "https://schema.org/" )
        SH            = Namespace( "http://www.w3.org/ns/shacl#" )
        SPEX          = Namespace( "https://purl.expasy.org/sparql-examples/ontology#" )
        XSD           = Namespace( "http://www.w3.org/2001/XMLSchema#" )
        OWL           = Namespace( "http://www.w3.org/2002/07/owl#" )
        g.bind( "rdf",    RDF )
        g.bind( "rdfs",   RDFS )
        g.bind( "rdfs",   RDFS )
        g.bind( "schema", SCHEMA )
        g.bind( "sh",     SH )
        g.bind( "spex",   SPEX )
        g.bind( "xsd",    XSD )
        g.bind( "owl",    OWL )
        g.bind( "sparql_query_" + config["server"]["repository"], SPARQLQUERY )
        g.bind( "prefix_" + config["server"]["repository"],    PREFIX )

        for path in config["queries"] :
            for dir, fn in expand_path( path, config["kgsteward_yaml_directory"] ):
                filename = dir + "/" + fn
                report( "read", filename )
                report( "parse file", filename )
                counter = counter + 1
                comment = []
                select  = [] # i.e. the SPARQL query itself
                name    = re.sub( r'(.*/|)([^/]+)\.\w+$', r'\2', filename )
                with open( filename ) as file:
                    for line in file:
                        if re.match( "^#", line ):
                            comment.append( re.sub( r"^#\s*", "", line.rstrip() ))
                        else:
                            select.append( line.rstrip().replace( "\t", "    "))
                            match = catch_key_value_rq.search( line )
                            if match:
                                prefix[match.group( 1 )] = match.group( 2 )
                # server.validate_sparql_query( "\n".join( select ), echo=args.v, timeout=args.timeout )
                iri = SPARQLQUERY[ "query_" + config["server"]["repository"] + str( counter ).rjust( 4, '0' )]
                g.add(( iri, RDF.type, SH.SPARQLExecutable ))
                g.add(( iri, RDF.type, SH.SPARQLSelectExecutable ))
                # g.add(( iri, RDFS.label,    Literal( "<b>" + name.replace( "_", " ") + "</b><br>")))
                g.add(( iri, RDFS.comment,  Literal( "<b>" + name.replace( "_", " ") + "</b><br>") + "\n".join( comment )))
                g.add(( iri, SH.prefixes,   prefixes_iri ))
                g.add(( iri, SH.select,     Literal( "\n".join( select ))))
                # g.add(( iri, SCHEMA.target, URIRef( server.get_endpoint_query()))) not portable
        if "prefixes" in config:
            for filename in config["prefixes"] :
                report( "parse file", filename )
                file = open( replace_env_var( filename ))
                for line in file:
                    match = catch_key_value_ttl.search( line )
                    if match:
                        prefix[ match.group( 1 ) ] = match.group( 2 )
        for key in prefix:
            g.add(( prefixes_iri,    SH.declare,   PREFIX[ key ]))
            g.add(( PREFIX[ key ], SH.prefix,    Literal( key )))
            g.add(( PREFIX[ key ], SH.namespace, Literal( prefix[key], datatype=XSD.anyURI )))
        g.serialize(
            format="turtle",
            destination = args.sib_swiss_editor
        )
        sys.exit( 0 )
        
    if args.x:
        if "queries" not in config:
            stop_error( "There are no 'queries' key in config! ")
        print_break()
        print_task( "Dump all query results in TSV format" )
        if not os.path.isdir( args.x ):
            stop_error( "Not a directory: " + args.x )
        for path in config["queries"] :
            for dir, fn in expand_path( path, config["kgsteward_yaml_directory"] ):
                filename = dir + "/" + fn
                print_break()
                report( "read file", filename )
                name    = re.sub( r'(.*/|)([^/]+)\.\w+$', r'\2', filename )
                with open( filename ) as file:
                    sparql = file.read()
                r = server.sparql_query( sparql, echo = args.v, timeout = args.timeout )
                if r is not None: # no timeout, i.e. the data are here
                    header, rows = sparql_result_to_table( r )
                    out_path =  args.x + "/" + name + ".tsv"
                    report( "write file", out_path )
                    with open( out_path, "w", encoding="utf-8" ) as f:
                        f.write( "\t".join( header ) + "\n" )
                        for row in sorted( rows ):
                            f.write( "\t".join( map( str, row )) + "\n" )

    if args.y:
        print_break()
        print_task( "Dump all contexts in TSV format" )
        if not os.path.isdir( args.y ):
            stop_error( "Not a directory: " + args.y )
        for item in config["dataset"]:
            print_break()
            header, rows = server.dump_context( item["context"], echo = args.v )
            out_path =  args.y + "/" +  item["name"] + ".tsv"
            report( "write file", out_path )
            with open( out_path, "w", encoding="utf-8"  ) as f:
                for row in sorted( rows ):
                    f.write( "\t".join( map( str, row )) + "\n" )

    # --------------------------------------------------------- #
    # Turn free access ON
    # --------------------------------------------------------- #

    if args.graphdb_free_access :
        if not config["server"]["brand"] == "graphdb":
            print_warn( "Option --graphdb_upload_queries not supported for server brand: " + config["server"]["brand"] )
        else:
            server.free_access()

    # --------------------------------------------------------- #
    # Print final repository status
    # FIXME: implement target_graph_context rewrite
    # --------------------------------------------------------- #

    config = update_config( server, config, echo = args.v )

    contexts = server.list_context()
    print_break()
    print_task( "Show current status" )
    print( colored("                            name        #triple        last modified    status", "blue" ))
    print( colored("================================        =======     =================== ======", "blue" ))
    for item in config["dataset"] :
        print( colored( '{:>32}   {:>12}    {:>20} {}'.format(
            item["name"],
            item["count"],
            item["date"],
            item["status"]
        ), "blue" ))
        context = item[ "context" ]
        if context in contexts :
            contexts.remove( context )
    for name in contexts:
        print( colored( '{:>32} : {:>12}    {:>20} {}'.format( name, "", "", "UNKNOWN" ), "blue" ))
    print_break()

    #  save_json_schema(  "doc/kgsteward.schema.json" )

# --------------------------------------------------------- #
# Main
# --------------------------------------------------------- #

if __name__ == "__main__":
    main()